import os

from depixelizer.io import read_png, get_writer
from .pixel_data import PixelData

class Depixelize():
    def __init__(self, filename, output_dir=None):
        self.filename = filename
        if output_dir:
            self.output_dir = output_dir
        else:
            self.output_dir = os.getcwd()
    
    def run(self):
        self.pixel_data = PixelData(read_png(self.filename))
        file_name = os.path.basename(self.filename).split(".")[0]
        self.pixel_data.depixelize()
        output_filepath = os.path.join(self.output_dir, file_name)
        writer = get_writer(self.pixel_data, output_filepath)
        writer.output_image()
        


from .bspline import Curve2Closed_BSpline
from .smoothen import smooth_spline

# Returns the slope of the line connecting two points
def slope(p0, p1):
    dx = p1[0] - p0[0]
    dy = p1[1] - p0[1]
    if dx == 0:
        return (
            dy * 99999999999999
        )  # Since dx is zero, we avoid division by zero. Instead, we multiply dy by a large number
    return 1.0 * dy / dx


class Path(object):
    # The path class encodes all the splines that are fit to a shape
    def __init__(self, shape_graph):
        self.path = self.make_path(shape_graph)
        self.shapes = set()

    def key(self):
        # This is the key used in the dict stored in the pixel graph
        return tuple(self.path)

    def make_path(self, shape_graph):
        # Creates the path and returns it
        nodes = set(shape_graph.nodes())
        path = [min(nodes)]
        neighbors = sorted(
            shape_graph.neighbors(path[0]), key=lambda p: slope(path[0], p)
        )
        path.append(neighbors[0])
        nodes.difference_update(
            path
        )  # This is equivalent to set difference between nodes and path

        # Go through the rest of the nodes and add nodes appropriately to the path
        p = path[-1]
        i = 0
        while nodes:
            for neighbor in shape_graph.neighbors(path[-1]):
                if neighbor in nodes:
                    nodes.remove(neighbor)
                    path.append(neighbor)
                    break
            if p != path[-1]:
                p = path[-1]
                i = 0
            else:
                i += 1
                if i == 3:
                    break
        return path

    def make_spline(self):
        # Fit a spline to the path
        self.spline = Curve2Closed_BSpline(self.path)

    def smooth_spline(self):
        # Optimise the spline that is already fit to the path
        self.smooth = smooth_spline(self.spline)



import random
from math import sin, cos, pi
from .point import Point


class SplineSmoother(
    object
):  # This class contains the functionality for optimization of the B-splines
    # Some constants to guide the optimization
    INTERVALS_PER_SPAN = 20
    POINT_GUESSES = 20
    GUESS_OFFSET = 0.05
    ITERATIONS = 20
    POSITIONAL_ENERGY_MULTIPLIER = 1

    def __init__(self, spline):
        # Initialization function and stores the spline to be optimized
        self.orig = spline
        self.spline = spline.copy()

    def E_C(self, index):
        # Energy due to curvature (formula directly used from paper)
        return self.spline.Energy_C(index, self.INTERVALS_PER_SPAN)

    def E_P(self, index):
        # Energy due to change in position (formula directly used from paper)
        orig = self.orig.points[index]
        point = self.spline.points[index]
        e_positional = abs(point - orig) ** 4
        return e_positional * self.POSITIONAL_ENERGY_MULTIPLIER

    def point_energy(self, index):
        # Total energy of a point = energy_curvature + energy_positional
        E1 = self.E_C(index)
        E2 = self.E_P(index)
        return E1 + E2

    def rand(self):
        # A random offset generator
        offset = random.random() * self.GUESS_OFFSET
        angle = random.random() * 2 * pi
        return offset * Point((cos(angle), sin(angle)))

    def smooth_point(self, index, start):
        # The function which is used to optimize the position of a point
        energies = [(self.point_energy(index), start)]
        for _ in range(
            self.POINT_GUESSES
        ):  # Around 20 guesses are made and the minimum energy one is chosen
            point = start + self.rand()
            self.spline.Move(index, point)
            energies.append((self.point_energy(index), point))
        self.spline.Move(index, min(energies)[1])  # Move the spline appropriately

    def smooth(self):
        # The function to smooth the spline
        for _it in range(self.ITERATIONS):
            for i, point in enumerate(self.spline.useful_points):
                self.smooth_point(i, point)


def smooth_spline(spline):
    # External function which creates an instance of the optimizer class
    smoother = SplineSmoother(spline)
    smoother.smooth()
    return smoother.spline



# Returns the tuple containing the vertices of the edge in sorted order
def get_sorted_vertices(edge):
    return tuple(sorted(edge[:2]))


# This function checks whether the given coordinates lie inside the image. If it is outside the dimensions of the image, then we return false.
def check_bounds(coordinate, size, offset=(0, 0)):
    x, y = map(
        sum, zip(coordinate, offset)
    )  # Add the coordinates of the given coordinate to the offset
    width, height = size
    return (
        0 <= x < width and 0 <= y < height
    )  # If the coordinates lie outside the dimensions of the image then return false


class Heuristics(object):
    sparse_window = (
        8,
        8,
    )  # The window around the edge which we examine for sparse heuristic

    def __init__(
        self, pixel_graph
    ):  # The pixel data of the image is loaded into the heuristic
        self.pixel_graph = pixel_graph

    def sparse_offset(
        self, edge
    ):  # This returns the boundaries of the window that must be examined in the sparse heuristic
        return (
            self.sparse_window[0] / 2 - 1 - min((edge[0][0], edge[1][0])),
            self.sparse_window[1] / 2 - 1 - min((edge[0][1], edge[1][1])),
        )

    def apply_heuristics(self, ambiguous_diagonal_pairs):
        # This function takes a list of ambiguous diagonal pairs and applies heuristics on them to resolve the diagonal
        # to remove and the diagonal to keep
        for edges in ambiguous_diagonal_pairs:
            self.weight_diagonals(
                *edges
            )  # Evaluate the heuristic value on each diagonal

        for edges in ambiguous_diagonal_pairs:
            min_weight = min(e[2]["weight"] for e in edges)
            for edge in edges:
                if edge[2]["weight"] == min_weight:
                    self.pixel_graph.remove_edge(
                        *edge[:2]
                    )  # Remove the edge which has minimum weight among the pair
                else:
                    edge[2].pop(
                        "weight"
                    )  # If it is not the diagonal removed, then the attribute weight is removed

    def weight_diagonals(self, edge1, edge2):
        # This function returns the weight for the pair of diagonals
        for edge in (edge1, edge2):
            self.weight_diagonal(edge)

    def weight_diagonal(self, edge):
        # This function returns the weight of the diagonal
        weights = [
            self.weight_curve(edge),
            self.weight_sparse(edge),
            self.weight_island(edge),
        ]
        edge[2]["weight"] = sum(weights)

    # This heuristic differs from the one in the paper slightly,
    # in the way that the paper suggests us to vote the longer edge with a weight equal to
    # difference in the lengths of the curves
    def weight_curve(self, edge):
        # This function returns the weight of the edge wrt the CURVE heuristic
        edges_in_the_curve = set([get_sorted_vertices(edge)])
        nodes_in_the_curve = list(edge[:2])

        while nodes_in_the_curve:
            node = nodes_in_the_curve.pop()
            edges = self.pixel_graph.edges(node, data=True)

            if (
                len(edges) != 2
            ):  # If the node is not a valence-2 node then it can't be part of a curve
                continue

            for edge in edges:
                edge = get_sorted_vertices(edge)
                if edge not in edges_in_the_curve:  # If edge not already seen
                    edges_in_the_curve.add(edge)  # Add the edge to the list
                    nodes_in_the_curve.extend(
                        n for n in edge if n != node
                    )  # Add the new nodes to the seen nodes
        return len(edges_in_the_curve)  # Return the length of the curve as the weight

    def weight_sparse(self, edge):
        # This function returns the weight of the edge wrt SPARSE heuristic
        nodes = list(edge[:2])
        nodes_in_the_sparse_window = set(nodes)
        offset = self.sparse_offset(edge)

        while nodes:
            node = nodes.pop()
            for n in self.pixel_graph.neighbors(node):
                if n in nodes_in_the_sparse_window:
                    continue
                if check_bounds(n, self.sparse_window, offset=offset):
                    nodes_in_the_sparse_window.add(n)
                    nodes.append(n)

        return -len(
            nodes_in_the_sparse_window
        )  # Returns the negative of the nodes found

    def weight_island(self, edge):
        # This function returns the weight of the edge wrt ISLAND heuristic
        # If any of teh two endpoints of the edge are valence-1 then removing the edge might
        # create an island
        if len(self.pixel_graph[edge[0]]) == 1 or len(self.pixel_graph[edge[1]]) == 1:
            return 5
        return 0



import png  # pypng


def read_png(filename):
    width, height, pixels, metadata = png.Reader(
        filename=filename
    ).asRGB8()  # Get the pixel data as RGB with 8 bits each
    pixel_data = []

    for pixel_row in pixels:  # For each row in the image
        pixel = []
        while pixel_row:
            pixel.append(
                (pixel_row.pop(0), pixel_row.pop(0), pixel_row.pop(0))
            )  # Get the RGB value of the pixel
        pixel_data.append(pixel)
    return pixel_data



class Shape(object):
    def __init__(self, pixels, value, corners):
        # This function initializes the Shape instance with pixel data, color and corners
        self.pixels = pixels
        self.value = value
        self.corners = corners
        self._outside_path = None  # The splines describing the boundary of the shape
        self._inside_paths = []  # The splines describing the inside of the shape

    # def _paths_attr(self, attr):
    # 	 paths = [list(reversed(getattr(self._outside_path, attr)))]
    # 	 paths.extend(getattr(path, attr) for path in self._inside_paths)

    # We define the following functions as property as we dont pass them any parameters and these are used in many
    # places. So, calling shape.paths is easier than shape.paths()
    @property
    def paths(self):
        # This function  returns the paths in the shape
        paths = [list(reversed(self._outside_path.path))]
        paths.extend(path.path for path in self._inside_paths)
        return paths

    @property
    def splines(self):
        # This function returns the splines in the shape
        paths = [self._outside_path.spline.reversed()]
        paths.extend(path.spline for path in self._inside_paths)
        return paths

    @property
    def smooth_splines(self):
        # This function returns the optimised splines in the shape
        paths = [self._outside_path.smooth.reversed()]
        paths.extend(path.smooth for path in self._inside_paths)
        return paths

    def add_outline(self, path, outside=False):
        # This function adds a boundary to the shape / an inside edge can also be added
        if outside:
            self._outside_path = path
        else:
            self._inside_paths.append(path)
        path.shapes.add(self)



import networkx
from math import sqrt

from .heuristic import Heuristics, check_bounds
from depixelizer.geometry import Shape, Path

# This generates all the coordinates from 0 to width and 0 to height i.e. it creates a walk over all pixels
def coordinates(size):
    # We use range as we don't want all the numbers from 0 to width in memory
    for y in range(size[1]):
        for x in range(size[0]):
            # We use yield because this number might be too large to store in memory. We are better off, generating it as we need them
            yield (x, y)


# Returns distance between two points
def distance(p0, p1):
    return sqrt((p1[0] - p0[0]) ** 2 + (p1[1] - p0[1]) ** 2)


# Removes the given elements from a collection of elements as a set
def remove(set_given, element):
    set_given.add(element)  # Add the element in case it isn't there
    # Remove both/the only copy of element in the set
    set_given.remove(element)


class PixelData(object):
    # The heuristics attribute of the pixeldata class
    HEURISTICS = Heuristics

    # Initialize with the pixel data
    def __init__(self, pixels):
        self.pixels = pixels
        self.width = len(pixels[0])
        self.height = len(pixels)
        self.size = (self.width, self.height)

    # The main function which contains all the procedures required for depixelizing the image
    def depixelize(self):

        self.create_pixel_graph()
        print("graph done")
        self.remove_diagonals()
        print("removed diagonals")
        self.create_grid_graph()
        print("grid graph done")
        self.deform_grid()
        print("deformed graph done")
        self.create_shapes()
        print("made shapes")
        self.get_boundaries()
        print("Obtained boundaries")
        self.add_shape_boundaries()
        print("shaped boundaries")
        self.smooth_splines()
        print("smoothed splines")

    def pixel(self, x, y):
        # The pixel data is loaded in the [y][x] format and not the [x][y]. Careful, while coding.
        return self.pixels[y][x]

    def create_grid_graph(self):
        # This function creates the grid graph of size (w+1)x(h+1) as described in the paper
        self.grid_graph = networkx.grid_2d_graph(self.width + 1, self.height + 1)

    def create_pixel_graph(self):
        # This function creates the pixel graph
        self.pixel_graph = networkx.Graph()
        # Add all the nodes and edges between them
        for x, y in coordinates(self.size):
            # Set the corners and connect the edges
            corners = set([(x, y), (x + 1, y), (x, y + 1), (x + 1, y + 1)])
            self.pixel_graph.add_node((x, y), value=self.pixel(x, y), corners=corners)
            self.add_pixel_edge((x, y), (x + 1, y))
            self.add_pixel_edge((x, y), (x, y + 1))
            self.add_pixel_edge((x, y), (x + 1, y - 1))
            self.add_pixel_edge((x, y), (x + 1, y + 1))

    def add_pixel_edge(self, pix0, pix1):
        # Adds a pixel edge in the graph iff it connects two similar colored pixels
        if check_bounds(pix1, self.size) and self.equal(pix0, pix1):
            attrs = {"diagonal": pix0[0] != pix1[0] and pix0[1] != pix1[1]}
            # Set the attributes appropriately
            self.pixel_graph.add_edge(pix0, pix1, **attrs)

    def equal(self, pix0, pix1):
        # Returns true if the color value in both pixels is same. Need to change this to have YUV channels.
        color0 = self.pixel(*pix0)
        color1 = self.pixel(*pix1)

        y0 = 0.299 * color0[0] + 0.587 * color0[1] + 0.114 * color0[2]
        u0 = 0.492 * (color0[2] - y0)
        v0 = 0.877 * (color0[0] - y0)

        y1 = 0.299 * color1[0] + 0.587 * color1[1] + 0.114 * color1[2]
        u1 = 0.492 * (color1[2] - y1)
        v1 = 0.877 * (color1[0] - y1)

        ydiff = abs(y0 - y1)
        udiff = abs(u0 - u1)
        vdiff = abs(v0 - v1)

        if ydiff > 48 or udiff > 7 or vdiff > 6:
            return False
        else:
            return True

        # return self.pixel(*pix0) == self.pixel(*pix1)

    def equal1(self, pix0, pix1):
        return self.pixel(*pix0) == self.pixel(*pix1)

    def remove_diagonals(self):
        # This function removes the diagonals from the graph
        ambiguous_diagonal_pairs = []

        for nodes in self.enumerate_blocks(2):
            edges = [
                e
                for e in self.pixel_graph.edges(nodes, data=True)
                if e[0] in nodes and e[1] in nodes
            ]  # Get all the edges which contain these nodes

            # Extract only the diagonals
            diagonals = [e for e in edges if e[2]["diagonal"]]
            if len(diagonals) == 2:
                if len(edges) == 6:
                    # We have a fully-connected block, so remove all diagonals.
                    for edge in diagonals:
                        self.pixel_graph.remove_edge(edge[0], edge[1])
                elif len(edges) == 2:
                    # We have an ambiguous pair to resolve.
                    ambiguous_diagonal_pairs.append(edges)
                else:
                    assert False, "Unexpected diagonal format"

        self.apply_diagonal_heuristics(ambiguous_diagonal_pairs)

    def apply_diagonal_heuristics(self, ambiguous_diagonal_pairs):
        self.HEURISTICS(self.pixel_graph).apply_heuristics(ambiguous_diagonal_pairs)

    def enumerate_blocks(self, size):
        # This function enumerates all the size x size blocks in a lazy manner
        for x, y in coordinates((self.width - size + 1, self.height - size + 1)):
            yield [(x + dx, y + dy) for dx in range(size) for dy in range(size)]

    def deform_grid(self):
        # This function deforms the grid as per the algorithm
        for node in self.pixel_graph.nodes():
            # Iterate through all the nodes in the graph and deform each of them
            self.deform_cell(node)

        # Collapse all valence-2 nodes to get a more smoother image
        removals = []
        for node in self.grid_graph.nodes():
            if node in ((0, 0), (0, self.size[1]), (self.size[0], 0), self.size):
                # Skip corner nodes of teh image as they shouldn't be collapsed, obviously!
                continue
            neighbors = list(self.grid_graph.neighbors(node))
            if (
                len(neighbors) == 2
            ):  # Connect the neighbors of the valence-2 node by an edge
                self.grid_graph.add_edge(*neighbors)
            if len(neighbors) <= 2:
                removals.append(node)

        # Remove all the valence-2 nodes
        for node in removals:
            self.grid_graph.remove_node(node)

        # Update pixel corner sets.
        for node, attrs in self.pixel_graph.nodes(data=True):
            corners = attrs["corners"]
            for corner in corners.copy():
                if corner not in self.grid_graph:
                    corners.remove(corner)

    def deform_cell(self, node):
        # This function deforms each cell as per the algorithm in the paper
        for neighbor in self.pixel_graph.neighbors(node):
            if node[0] == neighbor[0] or node[1] == neighbor[1]:
                # We only consider diagonals
                continue
            px_x = max(neighbor[0], node[0])
            px_y = max(neighbor[1], node[1])
            pixnode = (px_x, px_y)
            offset_x = neighbor[0] - node[0]
            offset_y = neighbor[1] - node[1]
            adj_node = (neighbor[0], node[1])
            if not self.equal1(node, adj_node):
                pn = (px_x, px_y - offset_y)
                mpn = (px_x, px_y - 0.5 * offset_y)
                npn = (px_x + 0.25 * offset_x, px_y - 0.25 * offset_y)
                remove(self.pixel_corners(adj_node), pixnode)
                self.pixel_corners(adj_node).add(npn)
                self.pixel_corners(node).add(npn)
                self.deform(pixnode, pn, mpn, npn)
            adj_node = (node[0], neighbor[1])
            if not self.equal1(node, adj_node):
                pn = (px_x - offset_x, px_y)
                mpn = (px_x - 0.5 * offset_x, px_y)
                npn = (px_x - 0.25 * offset_x, px_y + 0.25 * offset_y)
                remove(self.pixel_corners(adj_node), pixnode)
                self.pixel_corners(adj_node).add(npn)
                self.pixel_corners(node).add(npn)
                self.deform(pixnode, pn, mpn, npn)

    def pixel_corners(self, pixel):
        # This function returns the corners corresponding to the cell representing the given pixel
        return self.pixel_graph.nodes[pixel]["corners"]

    def deform(self, pixnode, pn, mpn, npn):
        # This function deforms the edge and connects other points which are offsetted by a constant from the
        # original position
        if mpn in self.grid_graph:
            self.grid_graph.remove_edge(mpn, pixnode)
        else:
            self.grid_graph.remove_edge(pn, pixnode)
            self.grid_graph.add_edge(pn, mpn)
        self.grid_graph.add_edge(mpn, npn)
        self.grid_graph.add_edge(npn, pixnode)

    def create_shapes(self):
        self.shapes = set()
        # Identify shapes in the graph by identifying the connected component subgraphs
        for pcg in (
            self.pixel_graph.subgraph(c).copy()
            for c in networkx.connected_components(self.pixel_graph)
        ):
            pixels = set()
            value = None
            corners = set()
            for pixel, attrs in pcg.nodes(data=True):
                pixels.add(pixel)
                corners.update(attrs["corners"])
                value = attrs["value"]
            self.shapes.add(Shape(pixels, value, corners))
            # Create a separate shape for each connected component subgraph and store all the pixels of the subgraph
            # in it

    def get_boundaries(self):
        # Remove internal edges from a copy of our pixgrid graph and just get the boundaries
        self.outlines_graph = networkx.Graph(self.grid_graph)
        for pixel, attrs in self.pixel_graph.nodes(data=True):
            corners = attrs["corners"]
            for neighbor in self.pixel_graph.neighbors(pixel):
                edge = corners & self.pixel_graph.nodes[neighbor]["corners"]
                if len(edge) != 2:  # If the number of edges is not 2
                    print(edge)
                # Remove the internal edges in the outlines graph
                elif self.outlines_graph.has_edge(*edge):
                    self.outlines_graph.remove_edge(*edge)
        for node in list(networkx.isolates(self.outlines_graph)):
            # Remove the nodes from the outline graph too
            self.outlines_graph.remove_node(node)

    def make_path(self, graph):
        # This function will create a path for a given subgraph
        path = Path(graph)
        key = path.key()
        if key not in self.paths:
            self.paths[key] = path
            path.make_spline()  # and also fit a spline to the path
        return self.paths[key]

    def add_shape_boundaries(self):
        self.paths = {}
        # Add the obtained boundaries to the corresponding shapes
        for shape in self.shapes:
            sg = self.outlines_graph.subgraph(shape.corners)
            for graph in (
                sg.subgraph(c).copy() for c in networkx.connected_components(sg)
            ):
                path = self.make_path(graph)
                if min(graph.nodes()) == min(sg.nodes()):
                    shape.add_outline(path, True)
                else:
                    shape.add_outline(path)

    def smooth_splines(self):
        # This function iterates through all the paths and tries to smooth each of them.
        print("Smoothing splines...")
        for i, path in enumerate(self.paths.values()):
            print(
                " * %s/%s (%s, %s)..."
                % (i + 1, len(self.paths), len(path.shapes), len(path.path))
            )
            if len(path.shapes) == 1:
                path.smooth = path.spline.copy()
                continue
            path.smooth_spline()  # Smooth each path which was previosuly fit with a spline



from math import sqrt
from .point import Point


class BSpline(object):
    # Class for B-Spline Curve
    def __init__(self, KnotVector, points, degree=None):
        # The initialization function which takes Knotvector, Control Points and degree(can be excluded) of the B-Spline
        self.KnotVector = tuple(KnotVector)
        self._points = [Point(p) for p in points]
        # Expected degree, if not specified in the constructor
        expected_degree = len(self.KnotVector) - len(self._points) - 1
        if degree is None:
            degree = expected_degree
        if degree != expected_degree:
            raise ValueError(
                "Degree expected is %s, got %s instead as Input."
                % (expected_degree, degree)
            )
        self.degree = degree
        self.remove_stored()

    def remove_stored(self):
        #   This removes the stored derivative value of the curve.
        self.storedvalue = {}

    def Move(self, i, value):
        #   This function moves a specific control point to a new location
        self._points[i] = value
        self.remove_stored()

    def __str__(self):
        return "<%s degree=%s, points=%s, KnotVector=%s>" % (
            type(self).__name__,
            self.degree,
            len(self.points),
            len(self.KnotVector),
        )

    def copy(self):
        #   Copy Constructor
        return type(self)(self.KnotVector, self.points, self.degree)

    # Few properties to ease syntax
    @property
    def domain(self):
        return (
            self.KnotVector[self.degree],
            self.KnotVector[len(self.KnotVector) - self.degree - 1],
        )

    @property
    def points(self):
        return tuple(self._points)

    @property
    def useful_points(self):
        return self.points

    # Overloading the call function of this class
    def __call__(self, u):
        # This function implements the DeBoor's algorithm
        s = len([uk for uk in self.KnotVector if uk == u])
        for k, uk in enumerate(self.KnotVector):
            if uk >= u:
                break
        if s == 0:
            k -= 1
        if self.degree == 0:
            if k == len(self.points):
                k -= 1
            return self.points[k]
        ps = [
            dict(
                zip(
                    range(k - self.degree, k - s + 1),
                    self.points[k - self.degree : k - s + 1],
                )
            )
        ]

        for r in range(1, self.degree - s + 1):
            ps.append({})
            for i in range(k - self.degree + r, k - s + 1):
                a = (u - self.KnotVector[i]) / (
                    self.KnotVector[i + self.degree - r + 1] - self.KnotVector[i]
                )
                ps[r][i] = (1 - a) * ps[r - 1][i - 1] + a * ps[r - 1][i]
        return ps[-1][k - s]

    def Quadratic_Bezier_Fit(self):
        # 	Finding a series of quadratic Beziers making up this spline.
        assert self.degree == 2
        control_points = self.points[1:-1]
        on_curve_points = [self(u) for u in self.KnotVector[2:-2]]
        ocp0 = on_curve_points[0]
        for cp, ocp1 in zip(control_points, on_curve_points[1:]):
            # Yield the starting point of the curve, control point and end point
            yield (ocp0.tuple, cp.tuple, ocp1.tuple)
            ocp0 = ocp1

    def Derivative(self):
        #   Returns the derivative of the curve, more Math!
        cached = self.storedvalue.get("1")
        if cached:
            return cached

        new_points = []
        p = self.degree
        for i in range(0, len(self.points) - 1):
            coeff = p / (self.KnotVector[i + 1 + p] - self.KnotVector[i + 1])
            new_points.append(coeff * (self.points[i + 1] - self.points[i]))

        cached = BSpline(self.KnotVector[1:-1], new_points, p - 1)
        self.storedvalue["1"] = cached
        return cached

    def Clamp(self, value):
        return max(self.domain[0], min(self.domain[1], value))

    def Span(self, index):
        return (
            self.Clamp(self.KnotVector[index]),
            self.Clamp(self.KnotVector[index + 1]),
        )

    def Points_In_Span(self, index):
        return [self.Span(index + i) for i in range(self.degree)]

    def Integrate_part(self, func, span, intervals):
        #   Integrates the given function on the Interval and returns value
        if span[0] == span[1]:
            return 0

        interval = (span[1] - span[0]) / intervals
        result = (func(span[0]) + func(span[1])) / 2
        for i in range(1, intervals):
            result += func(span[0] + i * interval)
        result *= interval

        return result

    def Integrate(self, index, func, intervals):
        # Integrates the function in intervals
        spans_ = self.Points_In_Span(index)
        spans = [span for span in spans_ if span[0] != span[1]]
        return sum(self.Integrate_part(func, span, intervals) for span in spans)

    def Curvature(self, u):
        #   Returns the curvature of the spline at a point
        d1 = self.Derivative()(u)
        d2 = self.Derivative().Derivative()(u)
        numerator = d1.x * d2.y - d1.y * d2.x
        denominator = sqrt(d1.x**2 + d1.y**2) ** 3
        if denominator == 0:
            return 0
        return abs(numerator / denominator)

    def Energy_C(self, index, intervals_per_span):
        #   Energy due to curvature at a point
        return self.Integrate(index, self.Curvature, intervals_per_span)

    def reversed(self):
        return type(self)(
            (1 - k for k in reversed(self.KnotVector)),
            reversed(self._points),
            self.degree,
        )


class Closed_BSpline(BSpline):
    # Class for representing closed b-pline curves where the start and edn point coincide
    def __init__(self, KnotVector, points, degree=None):
        super(Closed_BSpline, self).__init__(KnotVector, points, degree)
        self._unwrapped_len = len(self._points) - self.degree
        self.Wrap_check()

    def Wrap_check(self):
        # Checks the closed-ness of the spline
        if self._points[: self.degree] != self._points[-self.degree :]:
            raise ValueError("Points not wrapped at degree %s." % (self.degree,))

    def Move(self, index, value):
        # Moves the spline according to the change in position of a control point
        if not 0 <= index < len(self._points):
            raise IndexError(index)
        index = index % self._unwrapped_len
        super(Closed_BSpline, self).Move(index, value)
        if index < self.degree:
            super(Closed_BSpline, self).Move(index + self._unwrapped_len, value)

    @property
    def useful_points(self):
        return self.points[: -self.degree]

    def Span(self, index):
        def span(i):
            return (self.KnotVector[i], self.KnotVector[i + 1])

        d0, d1 = span(index)
        if d0 < self.domain[0]:
            d0, d1 = span(index + len(self.points) - self.degree)
        elif d1 > self.domain[1]:
            d0, d1 = span(index + self.degree - len(self.points))
        return self.Clamp(d0), self.Clamp(d1)


def Curve2Closed_BSpline(path, degree=2):
    # Returns the closed b-pline that is fit to the path given
    points = path + path[:degree]
    m = len(points) + degree
    KnotVector = [float(i) / m for i in range(m + 1)]

    return Closed_BSpline(KnotVector, points, degree)


def magnitude(point):
    return sqrt(point[0] ** 2 + point[2] ** 2)



import os.path
from math import sqrt
from svgwrite import Drawing


def distance(p0, p1):
    return sqrt((p1[0] - p0[0]) ** 2 + (p1[1] - p0[1]) ** 2)


def color(rgb_value_tuple):
    return "rgb(%s,%s,%s)" % rgb_value_tuple


def get_writer(data, filename):
    return SVGWriter(data, filename)


class SVGWriter(object):
    PIXEL_SCALE = 40
    CURVE_COLOR = (255, 127, 0)

    def __init__(self, data, filename, scale=None, color=None):
        self.name = filename
        self.pixel_data = data
        if scale:
            self.PIXEL_SCALE = scale
        if color:
            self.CURVE_COLOR = color

    def scale_pt(self, pt, offset=(0, 0)):
        return tuple(int((n + o) * self.PIXEL_SCALE) for n, o in zip(pt, offset))

    def output_image(self):
        filename = self.name
        # filename = filename+".svg"
        filename = os.path.join("outputs", "%s.%s" % (filename, "svg"))
        drawing = self.make_drawing(filename)
        self.draw_shapes(drawing)
        # self.insert_color(drawing)
        self.save_drawing(drawing, filename)

    def make_drawing(self, filename):
        return Drawing(filename)

    def save_drawing(self, drawing, filename):
        drawing.save()

    def draw_shapes(self, drawing):
        for shape in self.pixel_data.shapes:
            paths = getattr(shape, "smooth_splines")
            self.draw_spline(drawing, paths, shape.value)

    def draw_spline(self, drawing, splines, fill):
        if fill == (255, 255, 255):
            return
        path = []
        # points = []
        for spline in splines:
            curves = list(spline.Quadratic_Bezier_Fit())
            path.append("M")
            path.append(self.scale_pt(curves[0][0]))
            for curve in curves:
                path.append("Q")
                path.append(self.scale_pt(curve[1]))
                path.append(self.scale_pt(curve[2]))
                p0 = self.scale_pt(curve[0])
                p1 = self.scale_pt(curve[2])
            path.append("Z")
        drawing.add(drawing.path(path, stroke=color(fill), fill=color(fill)))



from depixelizer import Depixelize
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("image_path", help="Input Pixel Art Image")
parser.add_argument(
    "--output_dir",
    default=None,
    required=False,
    help="Output Dir, defaults to os.cwd()",
)


def depixelize_fn():
    args = parser.parse_args()
    if len(args.image_path) > 0:
        filename = args.image_path
    else:
        filename = input("Please enter the input image path")
    Depixelize(filename, args.output_dir).run()



from importlib.metadata import entry_points
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="depixelizer",
    version="0.0.1",
    author="Anirudh Vemula, Vamsidhar Yeddu",
    author_email="vvanirudh@gmail.com, vamsidhar666@gmail.com",
    description="An implementation of the paper Depixelizing Pixel Art by Kopf and Lischinski.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vvanirudh/Pixel-Art",
    project_urls={
        "Bug Tracker": "https://github.com/vvanirudh/Pixel-Art/issues",
    },
    entry_points={
        "console_scripts": ["depixelize=depixelizer.cmd:depixelize_fn"],
    },
    classifiers=["Programming Language :: Python :: 3"],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)



class Point(object):
    # A class to efficiently represent points. Value is stored as an imaginary number as it is easier in the caluculation relevant
    # to B-splines.
    def __init__(self, value):
        if isinstance(value, complex):
            self.value = value
        elif isinstance(value, (tuple, list)):
            self.value = value[0] + value[1] * 1j
        elif isinstance(value, Point):
            self.value = value.value
        else:
            pass

    # Some properties of the point
    @property
    def x(self):
        return self.value.real

    @property
    def y(self):
        return self.value.imag

    @property
    def tuple(self):
        return (self.x, self.y)

    def _op(self, op, other):
        if isinstance(other, Point):
            other = other.value
        return Point(getattr(self.value, op)(other))

    def __eq__(self, other):
        try:
            other = Point(other).value
        except ValueError:
            pass
        return self.value.__eq__(other)

    # Basic operators overloaded
    def __add__(self, other):
        return self._op("__add__", other)

    def __radd__(self, other):
        return self._op("__radd__", other)

    def __sub__(self, other):
        return self._op("__sub__", other)

    def __rsub__(self, other):
        return self._op("__rsub__", other)

    def __mul__(self, other):
        return self._op("__mul__", other)

    def __rmul__(self, other):
        return self._op("__rmul__", other)

    def __div__(self, other):
        return self._op("__div__", other)

    def __rdiv__(self, other):
        return self._op("__rdiv__", other)

    def __abs__(self):
        return abs(self.value)

    def round(self, places=5):
        return Point((round(self.x, places), round(self.y, places)))
